;; Generated package description from scroll-on-jump.el  -*- no-byte-compile: t -*-
(define-package "scroll-on-jump" "0.3.0.20260108.130950" "Scroll when jumping to a new point" '((emacs "26.2")) :commit "3e4e13fab04ab5acc7653d0474bf7b8b685d07d3" :authors '(("Campbell Barton" . "ideasman42@gmail.com")) :maintainer '("Campbell Barton" . "ideasman42@gmail.com") :url "https://codeberg.org/ideasman42/emacs-scroll-on-jump")
